/**
 * Copyright (c) 2009 Software Architecture Group
 *                    Hasso-Plattner-Institute, Potsdam, Germany
 *  
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */
package jcop.compiler;

import java.io.File;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;

import jcop.Globals;
import jcop.JCopAspect;
import jcop.Globals.CompilerOps;
import jcop.generation.AspectSourceCodeGeneration;
import jcop.generation.SourceCodeGenerator;
import jcop.generation.XMLOutlineGenerator;
import AST.BytecodeParser;
import AST.CompilationUnit;
import AST.Frontend;
import AST.JavaParser;
import AST.Program;
import AST.TypeDecl;

public class JCopCompiler extends Frontend {
	public static String concreteLayerName;
	public CompilerMessageStream msg;
	public AuxiliaryCompilerFilesHandler fileHandler;
	private long timestamp;
	private SourceCodeGenerator fileDumper;
	private XMLOutlineGenerator xmlOutlineGenerator;
	boolean layerClassGenerated = false;
	private CompilationUnit firstUnit;
	private HashSet<CompilationUnit> processedUnits;

	public static void main(String args[]) {
		if (!new JCopCompiler().compile(args))
			System.exit(1);	
	}

	public boolean compile(String args[]) {
		if (args.length == 0) {
			printUsage();
			return false;
		}	
		init(args);
		timestamp = System.currentTimeMillis();
		boolean compiled = process(args);
		
		if (compiled) {
			processConcreteLayers(firstUnit);	
			msg.print("compiled in " + (System.currentTimeMillis() - timestamp) + " millis\n");			 
			if (Globals.lang.equals("jcop"))
				fileHandler.writeJCopAspectToDisc(Program.hasOption("-debug"));
			fileHandler.cleanProgram();
		}	
		return compiled;
	}

	private void init(String[] args) {
		ArgumentParser argumentParser = new ArgumentParser(args);
		processedUnits = new HashSet<CompilationUnit>();		
		initOptions();		
		processArgs(argumentParser.getParsedArgs());
		msg = new CompilerMessageStream(Program.hasOption(CompilerOps.verbose) || Program.hasOption(CompilerOps.log));
		fileHandler = new AuxiliaryCompilerFilesHandler(msg);
		initArgs();
	}
	
	private void initArgs() {
		if (Program.hasOption(CompilerOps.debug))
			fileDumper = 
				new SourceCodeGenerator(getSourceCodeFolderName());
		if (Program.hasOption(CompilerOps.xmlOutlinePath))
			xmlOutlineGenerator = 
				new XMLOutlineGenerator(Program.getValueForOption(CompilerOps.xmlOutlinePath));				
	}
	
	private String getSourceCodeFolderName() {
		return fileHandler.getWorkingDir() + File.separator + "compiled-sources";
	}
	
	private boolean process(String[] args) {
		return process(args, new BytecodeParser(), new JavaParser() {
			public CompilationUnit parse(java.io.InputStream is, String fileName)
					throws java.io.IOException, beaver.Parser.Exception {
				return new parser.JavaParser().parse(is, fileName);
			}
		});
	}
		
	protected void processErrors(Collection errors, CompilationUnit unit) {		
		if (Program.hasOption(CompilerOps.xmlOutlinePath)) {
			System.err.println("xml");
			xmlOutlineGenerator.writeToFile(unit);
		}
			
		else {
			System.err.println("error");
			unit.transformation();
			unit.generateClassfile();
			//print("...error");
			
			super.processErrors(errors, unit);
		}
		  
	}
		
	protected void processNoErrors(CompilationUnit unit) {
		// processedUnits.add(unit.lookupType(Names.contextjPackage,
		// Names.layer));
		// ((CompilationUnit) concreteLayerDecl.getParent().getParent());
		// processedUnits.add(unit.lookupType(Names.contextjPackage,
		// Names.layer));
		
		TypeDecl layerType = unit.lookupType(Globals.jcopPackage, Globals.layer);
		CompilationUnit layerUnit = ((CompilationUnit) layerType.getParent().getParent());
		TypeDecl concreteLayerType = unit.lookupType(Globals.jcopPackage,
				Globals.concreteLayer);
		CompilationUnit concreteLayerUnit = ((CompilationUnit) concreteLayerType
				.getParent().getParent());

		if (!processedUnits.contains(unit) && (unit != concreteLayerUnit)
				&& (unit != layerUnit)) {
			
			if (Program.hasOption(CompilerOps.xmlOutlinePath))
				xmlOutlineGenerator.writeToFile(unit);
			
			unit.transformation();
			msg.print("compiling:" + unit.relativeName());
			if (firstUnit == null) {
				firstUnit = unit;
			}
			if (Program.hasOption(CompilerOps.debug))
				fileDumper.writeToFile(unit);
		
			unit.generateClassfile();
			msg.print(" ...done\n");
			//System.out.println("clear on exit");
			unit.clearOnExit();
			processedUnits.add(unit);
		}
	}

	private boolean existsConcreteLayer() {
		// return threadLocalConcreteLayerName.get() != null;
		return concreteLayerName != null;
	}

	private void processConcreteLayer(CompilationUnit unit, String pckg,
			String cls) {
		TypeDecl type = unit.lookupType(pckg, cls);
		CompilationUnit compilationUnit = ((CompilationUnit) type.getParent()
				.getParent());
		if (!processedUnits.contains(compilationUnit)) {

			compilationUnit.generateClassfile();

			if (Program.hasOption(("-debug")))
				fileDumper.writeToFile(compilationUnit);

			for (TypeDecl decl : compilationUnit.getTypeDecls()) {
				msg.print(" " + decl.getID());
				decl.transformation();
				decl.generateClassfile();
			}
			processedUnits.add(compilationUnit);
		}
	}

	protected void processConcreteLayers(CompilationUnit unit) {
		if (existsConcreteLayer()) {
			msg.print("compiling layers: ");
			processConcreteLayer(unit, Globals.jcopPackage, Globals.layer);
			processConcreteLayer(unit, Globals.jcopPackage,	Globals.concreteLayer);
			processConcreteLayer(unit, Globals.jcopPackage, concreteLayerName);
			msg.print(" ...done\n");
		}
	}

	@Override
	protected void initOptions() {
		super.initOptions();
		program.addKeyValueOption(CompilerOps.rootDir);
		program.addKeyValueOption(CompilerOps.layers);
		program.addKeyValueOption(CompilerOps.xmlOutlinePath);
		program.addKeyOption(CompilerOps.log);
		program.addKeyOption(CompilerOps.debug);		
	}

	@Override
	protected void printUsage() {
		System.out.println(
			name() + "\n"
			+ "Usage: " + Globals.version + " <options> <source files>\n"
			+ "where possible options include:\n"
			+ "  -layers <path>            Specify where to find class-in-layer source files\n"
			+ "  -sourcepath <path>        Specify where to find input source files. Only required if the sources\n" 
			+ "                            aren't located in the working directory.\n"					
			+ "  -xml-outline-path <path>  Generate an outline in XML format (useful for development environments).\n"
			+ "  -log                      Output JCop specific messages about what the compiler is doing\n"
			+ "  -verbose                  Output messages about what the compiler is doing\n"			
			+ "  -help                     Print a synopsis of standard options.\n"
			+ "  -debug                    Generates Java source files of the compiled classes into './compiled-sources/'.\n"
			+ "  -version                  Print version information.\n");
	}

	protected String url() {
		return "[http://hpi.uni-potsdam.de/swa]";
	}

	protected String name() {
		return Globals.lang + " Compiler";
	}

	protected String version() {
		return Globals.version;
	}
}
